# Copyright (c) 2015 Amazon.com, Inc. or its affiliates.  All Rights Reserved
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

import boto
from tests.compat import unittest
from boto.cloudhsm.exceptions import InvalidRequestException


class TestCloudHSM(unittest.TestCase):
    def setUp(self):
        self.cloudhsm = boto.connect_cloudhsm()

    def test_hapgs(self):
        label = 'my-hapg'
        response = self.cloudhsm.create_hapg(label=label)
        hapg_arn = response['HapgArn']
        self.addCleanup(self.cloudhsm.delete_hapg, hapg_arn)

        response = self.cloudhsm.list_hapgs()
        self.assertIn(hapg_arn, response['HapgList'])

    def test_validation_exception(self):
        invalid_arn = 'arn:aws:cloudhsm:us-east-1:123456789012:hapg-55214b8d'
        with self.assertRaises(InvalidRequestException):
            self.cloudhsm.describe_hapg(invalid_arn)
